@@ -239,6 +239,48 @@ class TestCallbackClient(unittest.TestCase, TestUtilMixin):
             [("!help ok 12", "")]
         )
 
+    def test_timeout(self):
+        """Test requests that timeout."""
+
+        replies = []
+        informs = []
+        timeout = 0.000001
+
+        def reply_cb(msg):
+            replies.append(msg)
+
+        def inform_cb(msg):
+            informs.append(msg)
+
+        self.client.request(
+            katcp.Message.request("help"),
+            reply_cb=reply_cb,
+            inform_cb=inform_cb,
+            timeout=timeout,
+        )
+
+        time.sleep(0.1)
+        self.assertEqual(len(replies), 1)
+        self.assertEqual(len(informs), 0)
+        self.assertEqual([msg.name for msg in replies], ["help"])
+        self.assertEqual([msg.arguments for msg in replies], [["fail", "Timed out after %f seconds" % timeout]])
+
+        del replies[:]
+        del informs[:]
+
+        # test next request succeeds
+        self.client.request(
+            katcp.Message.request("help"),
+            reply_cb=reply_cb,
+            inform_cb=inform_cb,
+        )
+
+        time.sleep(0.1)
+        self.assertEqual(len(replies), 1)
+        self.assertEqual(len(informs), 12)
+        self.assertEqual([msg.name for msg in replies + informs], ["help"]*len(replies+informs))
+        self.assertEqual([msg.arguments for msg in replies], [["ok", str(len(informs))]])
+
     def test_user_data(self):
         """Test callbacks with user data."""
         help_replies = []
@@ -309,7 +351,27 @@ class TestCallbackClient(unittest.TestCase, TestUtilMixin):
             replies, informs, done = results[thread_id]
             done.wait(1.0)
             self.assertEqual(len(replies), 1)
+            self.assertEqual(replies[0].arguments[0], "ok")
             if len(informs) != 12:
                 print thread_id, len(informs)
                 print [x.arguments[0] for x in informs]
             self.assertEqual(len(informs), 12)
+
+    def test_blocking_request(self):
+        """Test the callback client's blocking request."""
+        reply, informs = self.client.blocking_request(
+            katcp.Message.request("help"),
+        )
+
+        self.assertEqual(reply.name, "help")
+        self.assertEqual(reply.arguments, ["ok", "12"])
+        self.assertEqual(len(informs), 12)
+
+        reply, informs = self.client.blocking_request(
+            katcp.Message.request("help"),
+            timeout = 0.000001,
+        )
+
+        self.assertEqual(reply.name, "help")
+        self.assertEqual(reply.arguments[0], "fail")
+        self.assertTrue(reply.arguments[1].startswith("Timed out after"))
